package org.ow2.bonita.deployment;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import org.ow2.bonita.facade.def.element.BusinessArchive;
import org.ow2.bonita.facade.def.element.DeadlineDefinition;
import org.ow2.bonita.facade.def.element.HookDefinition;
import org.ow2.bonita.facade.def.majorElement.ActivityDefinition;
import org.ow2.bonita.facade.def.majorElement.ClientProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.DataFieldDefinition;
import org.ow2.bonita.facade.def.majorElement.ParticipantDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessFullDefinition;
import org.ow2.bonita.facade.def.majorElement.TransitionDefinition;
import org.ow2.bonita.facade.def.majorElement.ProcessDefinition.ProcessState;
import org.ow2.bonita.facade.exception.DeploymentException;
import org.ow2.bonita.facade.uuid.ActivityDefinitionUUID;
import org.ow2.bonita.facade.uuid.DataFieldDefinitionUUID;
import org.ow2.bonita.facade.uuid.ParticipantDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.TransitionDefinitionUUID;
import org.ow2.bonita.parsing.XpdlParser;
import org.ow2.bonita.pvm.internal.xml.Parse;
import org.ow2.bonita.services.util.ServiceEnvTool;
import org.ow2.bonita.util.ExceptionManager;
import org.ow2.bonita.util.Misc;

public class ProcessServerBuilder {

  public static ProcessFullDefinition buildServerProcess(BusinessArchive businessArchive) throws DeploymentException {
    ClientProcessDefinition clientProcess = null;
    if (businessArchive.isProcessFile()) {
      clientProcess = getProcessFromProcessFile(businessArchive);
    } else {
      clientProcess = getProcessFromXpdlFile(businessArchive);
    }
    
    ProcessFullDefinition process = (ProcessFullDefinition) clientProcess;
        
    final Date deployedDate = new Date();
    final String deployedBy = ServiceEnvTool.getUserId();
    final ProcessDefinitionUUID processUUID = ServiceEnvTool.getUUIDGenerator().getProcessDefinitionUUID(process.getName(), process.getVersion());
System.out.println(processUUID);
    process.setUUID(processUUID);
    
    process.setState(ProcessState.DEPLOYED);
    process.setDeployedDate(deployedDate);
    process.setDeployedBy(deployedBy);
    
    process.setBusinessArchive(businessArchive);
    Set<ParticipantDefinition> groups = process.getParticipants();
    if (groups != null) {
      for (ParticipantDefinition group : groups) {
        ParticipantDefinitionUUID uuid =
          ServiceEnvTool.getUUIDGenerator().getParticipantDefinitionUUID(processUUID,group.getName());
        group.setUUID(uuid);
        group.setProcessDefinitionUUID(processUUID);
      }
    }
    Set<DataFieldDefinition> data = process.getDataFields();
    if (data != null) {
      for (DataFieldDefinition d : data) {
        d.setProcessDefinitionUUID(processUUID);
        DataFieldDefinitionUUID uuid =
          ServiceEnvTool.getUUIDGenerator().getDataFieldDefinitionUUID(processUUID, d.getName());
        d.setUUID(uuid);
      }
    }
    Set<TransitionDefinition> transitions = process.getTransitions();
    if (transitions != null) {
      for (TransitionDefinition transition : transitions) {
        transition.setProcessDefinitionUUID(processUUID);
        TransitionDefinitionUUID uuid =
          ServiceEnvTool.getUUIDGenerator().getTransitionDefinitionUUID(processUUID, transition.getName());
        transition.setUUID(uuid);
      }
    }
    Set<ActivityDefinition> activities = process.getActivities();
    if (activities != null) {
      for (ActivityDefinition activity : activities) {
        activity.setProcessDefinitionUUID(processUUID);
        ActivityDefinitionUUID uuid =
          ServiceEnvTool.getUUIDGenerator().getActivityDefinitionUUID(processUUID, activity.getName());
        activity.setUUID(uuid);

        Set<DataFieldDefinition> activityData = activity.getDataFields();
        if (activityData != null) {
          for (DataFieldDefinition actData : activityData) {
            actData.setProcessDefinitionUUID(processUUID);
            DataFieldDefinitionUUID actDataUuid =
              ServiceEnvTool.getUUIDGenerator().getDataFieldDefinitionUUID(processUUID, actData.getName());
            actData.setUUID(actDataUuid);
          }
        }
      }
    }
    final Set<String> classDependencies = getClassDependencies(activities, groups);
    final Set<String> processDependencies = getProcessDependencies(activities);
    
    process.setClassDependencies(classDependencies);
    process.setProcessDependencies(processDependencies);
    
    return process;
  }
  

  protected static Set<String> getProcessDependencies(final Set<ActivityDefinition> activities) {
    Set<String> dependencies = new HashSet<String>();
    if (activities != null) {
      for (final ActivityDefinition activity : activities) {
        if (activity.getSubFlow() != null) {
          final String subProcessId = activity.getSubFlow().getProcessId();
          //if (!containsProcess(processes, subProcessId)) {
            ProcessFullDefinition subProcess =
              ServiceEnvTool.getJournalQueriers().getLastDeployedProcess(activity.getSubFlow().getProcessId(),
                  ProcessDefinition.ProcessState.DEPLOYED);
            if (subProcess == null) {
              final Set<ProcessFullDefinition> prs = ServiceEnvTool.getJournalQueriers().getProcesses(
                  activity.getSubFlow().getProcessId(), ProcessDefinition.ProcessState.DEPLOYED);
              if (prs != null && !prs.isEmpty() && prs.size() >= 1) {
                subProcess = prs.iterator().next();
              }
            }
            if (subProcess == null) {
              throw new DeploymentRuntimeException("Unable to find process \"" + subProcessId + "\" used in subFlow. "
                  + "If the subFlow process is in an external package, "
                  + "please check that this package has already been deployed");
            }
            final ProcessDefinition processDef =
              ServiceEnvTool.getJournalQueriers().getProcess(subProcess.getUUID());
            if (processDef.getUndeployedBy() != null) {
              throw new DeploymentRuntimeException("Unable to find process \"" + subProcessId + "\" used in subFlow. ");
            }
            dependencies.add(processDef.getUUID().toString());
          //}
        }
      }
    }
    if (dependencies.isEmpty()) {
      dependencies = null;
    }
    return dependencies;
  }
  
  private static Set<String> getClassDependencies(final Set<ActivityDefinition> activities, final Set<ParticipantDefinition> participants) {
    Set<String> dependencies = new HashSet<String>();
    dependencies.addAll(getParticipantsDependencies(participants));
    dependencies.addAll(getActivitiesDependencies(activities));
    if (dependencies.isEmpty()) {
      dependencies = null;
    }
    return dependencies;
  }
  
  private static Set<String> getActivitiesDependencies(final Set<ActivityDefinition> activities) {
    final Set<String> dependencies = new HashSet<String>();
    if (activities != null) {
      for (final ActivityDefinition activity : activities) {
        if (activity.getPerformerAssign() != null) {
          dependencies.add(activity.getPerformerAssign().getClassName());
        }
        if (activity.getHooks() != null) {
          for (final HookDefinition hook : activity.getHooks()) {
            dependencies.add(hook.getClassName());
          }
        }
        if (activity.getDeadlines() != null) {
          for (final DeadlineDefinition deadline : activity.getDeadlines()) {
            dependencies.add(deadline.getExceptionName());
          }
        }
        if (activity.getMultiInstantiationDefinition() != null) {
          dependencies.add(activity.getMultiInstantiationDefinition().getClassName());
        }
      }
    }
    return dependencies;
  }
  
  private static Set<String> getParticipantsDependencies(final Set<ParticipantDefinition> participants) {
    final Set<String> dependencies = new HashSet<String>();
    if (participants != null) {
      for (final ParticipantDefinition participant : participants) {
        if (participant.getRoleMapper() != null) {
          dependencies.add(participant.getRoleMapper().getClassName());
        }
      }
    }
    return dependencies;
  }
  
  private static ClientProcessDefinition getProcessFromProcessFile(final BusinessArchive businessArchive) throws DeploymentException {
    ClientProcessDefinition process = null;
    final byte[] def = businessArchive.getProcessFile();
    if (def == null) {
      String message = ExceptionManager.getInstance().getFullMessage("bd_D_1");
      throw new DeploymentException(message);
    }
    try {
      ObjectInputStream  ois = new ObjectInputStream(new ByteArrayInputStream(def));
      process = (ProcessFullDefinition) ois.readObject();
    } catch (ClassNotFoundException e) {
      throw new DeploymentException(e.getMessage());
    } catch (IOException e1) {
      throw new DeploymentException(e1.getMessage());
    }

    Misc.badStateIfNull(process, "Ouch! The returned ProcessFullDefinition is null!");
    return process;
  }

  private static ClientProcessDefinition getProcessFromXpdlFile(final BusinessArchive businessArchive) throws DeploymentException {
    ClientProcessDefinition process = null;
    final Parse parse = new XpdlParser().createParse();
    parse.pushObject(businessArchive);

    // now the xpdl process will be parsed.
    final byte[] xpdl = businessArchive.getProcessFile();
    if (xpdl == null) {
      String message = ExceptionManager.getInstance().getFullMessage("bd_D_1");
      throw new DeploymentException(message);
    }

    // make the xpdl package is available to the xpdl parser by pushing it
    // on the object stack in the parse.
    parse.setInputStream(new ByteArrayInputStream(xpdl));
    process = (ClientProcessDefinition) parse.execute().getDocumentObject();

    //CHECK EVERYTHING IS OK
    parse.checkProblems("xpdl file");

    parse.popObject();
    Misc.badStateIfNull(process, "Ouch! The returned ProcessFullDefinition is null!");
    return process;
  }
}
